#!/usr/bin/env python

# $Id: tm.py,v 1.7 2001/05/31 12:17:50 larsga Exp $

# ----------------------------------------------------------------------------
#
# tmproc - a Topic Map engine
#
# Copyright (C) 1999, 2000 Geir O. Grnmo <grove@ontopia.net>
#
# ----------------------------------------------------------------------------

"""A collection of in-memory Topic Map classes."""

__author__ = "Geir Ove Grnmo, grove@ontopia.net"
__version__ = "0.21dev"

# ----------------------------------------------------------------------------
# TODO:
#
# Documentation:
#
# - Add docstrings to all classes and their methods 
# - Make Pythondoc documentation
# - Document the interfaces.

# General:
#
# - Add copyright notice to all modules/files.
# - Synchronize the interfaces with tmdb.py
# - sgml_id on Association and Facet?
# - Use sgml id/idref notation as the default loctype on Locations
# - Note that Locations may point to Topics, and usually do.
# - Store the system identifier of the topic map document - in order
#   to be able to locate the origin of the tm objects.
# - Authority collections [related to public subjects]?
# - .scope = [] ~ the unconstrained scope
# - add a TMConfigurer class.
# - reflexive and transitive associations
#
# + rename .get_public_subject() to get_identity() [fixed 990301]
# + Add .id and get_id() on all tm objects [fixed 990301]
# + References should be resolved using objects, not sgml_id. this should be
#   resolved while loading. [Fixed 990222, grove]
#
# Testing:
#
# - Make more sample documents
# - Make regression test
#
# Questions:
#
# - Should all mnemonics use the get_mnemnic() method instead of 
#   get_linktype(), get_rolename() etc.?
# - Should objects known about their parents [add a pointer to the parent]?
# - Should exceptions be thrown when nonexisting items are tried removed?
# - Should exceptions be thrown when already existing items are tried added?
# - Is the original SGML-id really needed in a topic map system?
# - Store topic characteristics etc. based on its scope -- using a dictionary

# ============================================================================
# Imported modules
# ============================================================================

import string

# ============================================================================
# Topic Map Factory
# ============================================================================

class TopicMapFactory:

    """Topic Map Facory. Used to make various tm objects."""

    # -----------------------------------------------------------------------------
    # Initialization
    # -----------------------------------------------------------------------------

    def __init__(self, module_keywords):

	self.__args = module_keywords
	
	self.last_topic_map = None
	self.last_topic = None
	self.last_name = None
	self.last_topic_name = None
	self.last_occurrence = None
	self.last_association = None
	self.last_association_role = None
	self.last_facet = None
	self.last_facet_value = None
	self.last_locator = None

    # -----------------------------------------------------------------------------
    # Factory methods
    # -----------------------------------------------------------------------------

    def make_topic_map(self, name="", addthems=()):
	
	"""Make a new TopicMap object."""

	self.last_topic_map = TopicMap(name, addthems)
	return self.last_topic_map
    
    def make_topic(self, tm, sgml_id="", identity="", linktype="", types=(), scopes=(), names=(), occurs=()):

	"""Make a new Topic object."""

	self.last_topic = tm.get_topic_by_sgml_id(sgml_id)
	if not self.last_topic:
	    self.last_topic = Topic(sgml_id, identity, linktype, types, scopes, names, occurs)
	    tm.add_topic(self.last_topic)
	return self.last_topic

    def make_name(self, tm, name="", scope=()):

	"""Make a new TopicName object."""

	self.last_name = Name(name, scope)
	return self.last_name

    def make_topic_name(self, tm, scopes=(), basenames=(), dispnames=(), sortnames=()):

	"""Make a new TopicName object."""

	self.last_topic_name = TopicName(scopes, basenames, dispnames, sortnames)
	return self.last_topic_name

    def make_occurrence(self, tm, rolename="", type=None, scopes=(), locators=()):

	"""Make a new Occurrence object."""

	self.last_occurrence = Occurrence(rolename, type, scopes, locators)
	return self.last_occurrence

    def make_association(self, tm, linktype="", roles=(), scopes=(), type=None):

	"""Make a new Association object."""

	self.last_association = Association(linktype, type, roles, scopes)
	tm.add_association(self.last_association)
	return self.last_association

    def make_association_role(self, tm, rolename="", type=None, locators=()):

	"""Make a new AssociationRole object."""

	self.last_association_role = AssociationRole(rolename, type, locators)
	return self.last_association_role

    def make_facet(self, tm, linktype="", type=None, values=()):

	"""Make a new Facet object."""

	self.last_facet = Facet(linktype, type, values)
	tm.add_facet(self.last_facet)
	return self.last_facet

    def make_facet_value(self, tm, value="", type=None):

	"""Make a new FacetValue object."""

	self.last_facet_value = FacetValue(value, type)
	return self.last_facet_value

    def make_locator(self, tm, address="", loctype="sgml"):

	"""Make a new FacetValue object."""

	if not loctype: loctype = "sgml"
	self.last_locator = Locator(address, loctype)
	return self.last_locator
        
# ============================================================================
# Topic Map classes
# ============================================================================

class TopicMap:

    """5.1 Topic Map"""

    # -----------------------------------------------------------------------------
    # Initialization
    # -----------------------------------------------------------------------------

    def __init__(self, name="", addthems=()):
	
	self.__name = name                # String
	self.__addthems = list(addthems)  # Topic*

	self.__topics = {}                # Topic*
	self.__associations = {}          # Association*
	self.__facets = {}                # Facet*
	
    def __str__(self):
	return self.get_name()

    def __getitem__(self, sgml_id):

	"""Get topic by sgml id."""
	
	return self.get_topic_by_sgml_id(sgml_id)
    
    # -----------------------------------------------------------------------------
    # Finalizer method
    # -----------------------------------------------------------------------------

    def finalize(self):

	"""Finalize the TopicMap object and resolve any missing information."""

	pass
    
    # -----------------------------------------------------------------------------
    # Getters
    # -----------------------------------------------------------------------------

    def get_id(self):

	"""Get the internal TopicMap id."""
	
	return hash(self)

    def get_name(self):

	"""Get the name of the topic map."""
	
	return self.__name

    def get_added_themes(self):

	"""Get the added themes of the topic map."""
	
	return self.__addthems

    def has_topic_by_sgml_id(self, sgml_id):

	"""Does a topic with the specified sgml-id exist in this topic map?"""

	return dict_has(self.__topics, "get_sgml_id", (), sgml_id)

    def has_topic(self, topic):

	"""Does the topic exist in this topic map?"""

	return self.__topics.has_key(topic)

    def has_association(self, association):

	"""Does the association exist in this topic map?"""

	return self.__associations.has_key(association)

    def has_facet(self, facet):

	"""Does the facet exist in this topic map?"""

	return self.__facets.has_key(facet)

    # -----------------------------------------------------------------------------
    # Setters
    # -----------------------------------------------------------------------------

    def set_name(self, name):

	"""Set the name of the topic map."""
	
	self.__name = name

    def set_added_themes(self, themes):

	"""Set the added themes of the topic map."""

	for theme in themes:
	    self.add_addthems(theme)

    def add_added_themes(self, theme):

	"""Set the added theme of the topic map."""
	
	self.__addthems.append(theme)

    def add_topic(self, topic):

	"""Add the topic to this topic map."""

	if self.__topics.has_key(topic):
	    raise "The topic %s already exist in the topic map." % (sgml_id)
	self.__topics[topic] = None

    def remove_topic(self, topic):

	"""Remove the topic from this topic map."""

	# FIXME: Must check all other references before deleting.
	remove_from_dict(self.__topics, topic, "The topic map doesn't have that topic.")
	
    def add_association(self, association):

	"""Add the association to this topic map."""

	self.__associations[association] = None
	
    def remove_association(self, association):

	"""Remove the association from this topic map."""

	remove_from_dict(self.__associations, association, "The topic map doesn't have that association.")

    def add_facet(self, facet):

	"""Add the facet to this topic map."""

	self.__facets[facet] = None

    def remove_facet(self, facet):

	"""Remove the facet from this topic map."""

	remove_from_dict(self.__facets, facet, "The topic map doesn't have that facet.")

    # -----------------------------------------------------------------------------
    # Topic access
    # -----------------------------------------------------------------------------

    def __getitem__(self, sgml_id):

	"""Get topic by sgml id."""

	return self.get_topic_by_sgml_id(sgml_id)
    
    def get_topic_by_sgml_id(self, sgml_id):

	"""Get topic by id. Returns a single topic."""

	return dict_has(self.__topics, "get_sgml_id", (), sgml_id)
    
    # =============================================================================
    # Topic map objects
    # =============================================================================
    
    # Topics
    def get_topics(self):

	"""Get a list containing all topics."""
	
	return self.__topics.keys()
    
    # Associations
    def get_associations(self):

	"""Get a list containing all associations."""
	
	return self.__associations.keys()

    # Facets
    def get_facets(self):

	"""Get a list containing all facets."""

	return self.__facets.keys()

    # =============================================================================
    # Type information
    # =============================================================================
	
    def get_types_list(self, tm_object, transitive=0):

	"""Get a list containing all topic types connected to the specified topic map object.

	This is a flattened version of the get_types_hierarchy method or the keys
	of the dictionary returned by the get_types_dict method."""
	
	# Return the keys of the get_types_dict method
	return self.__get_types_dict(tm_object, transitive).keys()
    
    def __get_types_dict(self, tm_object, transitive=0, types=None):

	"""Get dictionary containing all topic types connected to the specified topic map object.

	The keys are the topics themselves, the values are None."""

       	if types == None: types = {}
	
	# Loop over all specified types
	if hasattr(tm_object, "get_types"):
	    _types = tm_object.get_types()
	elif hasattr(tm_object, "get_type"):
	    _types = (tm_object.get_type(),)

	for type in _types:
	    # Is type already registered?
	    if types.has_key(type): continue
	    # Is this an unknown topic type
	    types[type] = None
	    if transitive:
		# Get types of supertype
		self.__get_types_dict(type, 1, types)
		
	# Return a pointer to the types dictionary
	return types
    
    def get_types_hierarchy(self, tm_object, types=None):

	"""Get nested list containing a hierarchy of types connected to this topic map object.

	Note that this methods traverses the types tree recursively to collect
	all supertypes. """
	
	if types == None: types = {}
      	# Create a new type list
	type_list = []
	# Add this topic to the types list and the types dictionary
	type_list.append(tm_object)
	types[tm_object] = None
	# Loop over all specified types
        for type in tm_object.get_types():
	    # Is type already registered, or is this an unknown topic type?
	    if types.has_key(type):
		type_list.append([type])
		continue
	    # Get types of supertypes
	    _types = self.get_types_hierarchy(type, types)
	    # Add supertypes to types list
	    type_list.append(_types)
	# Return a pointer to the types list
	return type_list

    def is_type(self, tm_object, type, transitive=0):

	"""Is the specified topic map object of this type?

	Note that if transitive in specified this methods traverses
	the types tree recursively to compare the specified type
	against all supertypes."""
	
	# Does the types dict contain this type?
	if self.__get_types_dict(tm_object, transitive).has_key(type):
	    return 1
	return 0
    
    def get_objects_of_type(self, type, tm_objects=None, transitive=0):

	"""Get a list containing all topic map objects that is of a given type.

	If tm_objects is None, all objects in the topic map is evaluated.
	
	transitive = 0 means that only direct types should be
	evaluated, transitive=1 means that all transitive types
	should be evaluated."""
	
	of_type = []
	if tm_objects == None: tm_objects = self.__get_objects()
	
	# Loop over objects
	for tm_object in tm_objects:
	    if type in self.get_types_list(tm_object, transitive):
		if type != tm_object:
		    of_type.append(tm_object)

	return of_type

    def get_object_types(self, tm_objects=None, transitive=0):

	"""Get a list containing the types describing the given topic map objects.
	
	If tm_objects is None, all objects in the topic map is evaluated."""

	types = {}
	if tm_objects == None: tm_objects = self.__get_objects()
	    
	# Loop over topics
	for tm_object in tm_objects:
	    # Loop over the types of the tm_object
	    for type in self.get_types_list(tm_object, transitive):
		types[type] = None

	return types.keys()

    def __get_objects(self):

	"""Get a list of all topic map objects (topics, associations
	and facets) that belong to the topic map."""
	
	return self.get_topics() + self.get_associations() + self.get_facets()
    
    # -----------------------------------------------------------------------------
    # Topic types
    # -----------------------------------------------------------------------------

    def get_topics_of_type(self, type, topics=None, transitive=0):

	"""Get a list containing all topics that is of a given topic type.

	If topics is None, all topics in the topic map is evaluated.
	
	transitive = 0 means that only direct types should be
	evaluated, transitive=1 means that all transitive topic types
	should be evaluated."""

	if topics == None: topics=self.get_topics()
	return self.get_objects_of_type(type, topics, transitive)
    
    def get_topic_types(self, topics=None, transitive=0):

	"""Get a list containing the topic types describing the given topics.
	
	If topics is None, all topics in the topic map is evaluated."""

	if topics == None: topics=self.get_topics()
	return self.get_object_types(topics, transitive)
    
    # -----------------------------------------------------------------------------
    # Association types
    # -----------------------------------------------------------------------------

    def get_associations_of_type(self, type, associations=None, transitive=0):

	"""Get a list containing all associations that is of a given association type.

	If associations is None, all associations in the topic map is evaluated."""
	
	if associations == None: associations=self.get_associations()
	return self.get_objects_of_type(type, associations, transitive)

    def get_association_types(self, associations=None, transitive=0):

	"""Get a list containing the topic types describing the given associations.

	If associations is None, all associations in the topic map is evaluated."""

	if associations == None: associations=self.get_associations()
	return self.get_object_types(associations, transitive)

    # FIXME: do nifty queries on associations. Lots of incredibly cool thing my be done here.
    
    # -----------------------------------------------------------------------------
    # Facet types
    # -----------------------------------------------------------------------------

    def get_facets_of_type(self, type, facets=None, transitive=0):

	"""Get a list containing all facets that is of a given facet type.

	If facets is None, all facets in the topic map is evaluated."""
	
	if facets == None: facets=self.get_facets()
	return self.get_objects_of_type(type, facets, transitive)

    def get_facet_types(self, facets=None, transitive=0):

	"""Get a list containing the topic types describing the given facets.

	If facets is None, all facets in the topic map is evaluated."""
	
	if facets == None: facets=self.get_facets()
	return self.get_object_types(facets, transitive)

    # =============================================================================
    # Scope information
    # =============================================================================

    def get_scopes_list(self, tm_object, transitive=0):

	"""Get a list containing all scopes connected to the specified topic map object."""
	
	# Return the keys of the get_scopes_dict method
	return self.__get_scopes_dict(tm_object, transitive).keys()
    
    def __get_scopes_dict(self, tm_object, transitive=0, scopes=None):

	"""Get dictionary containing all scopes connected to the specified topic map object.

	The keys are the topics themselves, the values are None."""

       	if scopes == None: scopes = {}
	
	# Loop over all specified scopes
	for scope in tm_object.get_scopes():
	    # Is scope already registered?
	    if scopes.has_key(scope): continue
	    # Is this an unknown topic scope
	    scopes[scope] = None
	    if transitive:
		# Get scopes of superscope
		self.__get_scopes_dict(scope, 1, scopes)
		
	# Return a pointer to the scopes dictionary
	return scopes
    
    def get_objects_in_scope(self, scope, tm_objects=None, transitive=0):

	"""Get a list containing all topic map objects that is in the given scope.

	If tm_objects is None, all objects in the topic map is evaluated."""
	
	in_scope = []
	if tm_objects == None: tm_objects = self.__get_objects()
	
	# Loop over objects
	for tm_object in tm_objects:
	    if scope in self.get_scopes_list(tm_object, transitive):
		in_scope.append(tm_object)
		
	return in_scope

    def get_object_scopes(self, tm_objects=None, transitive=0):

	"""Get a list containing the scopes of the given topic map objects.
	
	If tm_objects is None, all objects in the topic map is evaluated."""

	scopes = {}
	if tm_objects == None: tm_objects = self.__get_objects()
	    
	# Loop over topics
	for tm_object in tm_objects:
	    # Loop over the types of the tm_object
	    for scope in self.get_scopes_list(tm_object, transitive):
		scopes[scope] = None

	return scopes.keys()
    
    # -----------------------------------------------------------------------------
    # Association scopes
    # -----------------------------------------------------------------------------

    def get_associations_in_scope(self, scope, associations=None, transitive=0):

	"""Get a list containing all associations that is of a given association type.

	If associations is None, all associations in the topic map is evaluated."""
	
	if associations == None: associations=self.get_associations()
	return self.get_objects_in_scope(scope, associations, transitive)

    def get_association_scopes(self, associations=None, transitive=0):

	"""Get a list containing the scopes that the given associations exists in.

	If associations is None, all associations in the topic map is evaluated."""

	if associations == None: associations=self.get_associations()
	return self.get_object_scopes(associations, transitive)

    # -----------------------------------------------------------------------------
    # Topic relations
    # -----------------------------------------------------------------------------

    def get_related_associations(self, topic, associations=None):

	"""Get a list of associations where the topic is involved.

	If associations is None, all associations in the topic map is evaluated."""

	if associations == None: associations=self.get_associations()
	rel_assoc = {}
	for assoc in associations:
	    for assoc_role in assoc.get_roles():
		for topic_loc in assoc_role.get_locators():
		    if  topic == topic_loc:
			rel_assoc[assoc] = None

	return rel_assoc.keys()

# =============================================================================

class Topic:

    """5.2 Topic link"""
    
    # -----------------------------------------------------------------------------
    # Initialization
    # -----------------------------------------------------------------------------

    def __init__(self, sgml_id="", identity="", linktype="", types=(), scopes=(), names=(), occurs=()):

	self.__sgml_id = sgml_id           # String
	self.__identity = identity         # String
	self.__linktype = linktype         # String
	self.__types = list(types)         # Topic+
	self.__scopes = list(scopes)       # Topic+
	self.__names = list(names)         # TopicName+
	self.__occurrences = list(occurs)  # Occurrence*

	self.__associations = []

    def __str__(self):
	# return '<tm.Topic %d/%s>' % (self.get_id(), self.__sgml_id)
	return '<tm.Topic %s>' % (self.__sgml_id)

    def __repr__(self):
	return self.__str__()

    # -----------------------------------------------------------------------------
    # Finalizer method
    # -----------------------------------------------------------------------------

    def finalize(self):

	"""Finalize the Topic object and resolve any missing information."""

	pass

    # -----------------------------------------------------------------------------
    # Getters
    # -----------------------------------------------------------------------------

    def get_id(self):

	"""Get the internal Topic id."""
	
	return hash(self)

    def get_sgml_id(self):

	"""Get topic id."""
	
	return self.__sgml_id

    def get_identity(self):

	"""Get the topic's subject identity."""
	
	return self.__identity

    def get_linktype(self):

	"""Get the topic link type."""
	
	return self.__linktype

    def get_names(self):

	"""Get list of all names."""
	
	return self.__names

    def get_occurrences(self):

	"""Get list of occurrences."""
	
	return self.__occurrences

    def get_types(self):

	"""Get list of topic types."""
	
	return self.__types

    def get_scopes(self):

	"""Get the scopes that apply to this topic."""
	
	return self.__scopes

    # -----------------------------------------------------------------------------
    # Setters
    # -----------------------------------------------------------------------------

    def set_sgml_id(self, sgml_id):

	"""Set topic id."""
	
	self.__sgml_id = sgml_id

    def set_identity(self, identity):

	"""Set the subject identity."""
	
	self.__identity = identity

    def set_linktype(self, linktype):

	"""Set the topic link type."""
	
	self.__linktype = linktype

    def set_names(self, names):

	"""Set list of all names."""

	for name in names:
	    self.add_name(name)

    def add_name(self, name):

	"""Add name to the list of names."""
	
	self.__names.append(name)

    def set_occurrences(self, occurrences):

	"""Set list of occurrences."""

	for occurrence in occurrences:
	    self.add_occurrence(occurrence)

    def add_occurrence(self, occurrence):

	"""Add occurrence to the list of occurrences."""
	
	self.__occurrences.append(occurrence)

    def set_types(self, types):

	"""Set list of topic types."""

	for type in types:
	    self.add_type(type)

    def add_type(self, type):

	"""Add type to the list of topic types."""
	
	self.__types.append(type)

    def set_scopes(self, scopes):

	"""Set list of scopes."""

	for scope in scopes:
	    self.add_scope(scope)

    def add_scope(self, scope):

	"""Add scope to the list of scopes."""
	
	self.__scopes.append(scope)

    # -----------------------------------------------------------------------------
    # Associations [Experimental]
    # -----------------------------------------------------------------------------

    def get_associations(self):

	"""Get the list of associations that this topic is an anchor of."""
	
	return self.__associations

    def set_associations(self, associations):

	"""Set the list of associations that this topic is an anchor of."""

	for association in associations:
	    self.add_association(association)

    def add_association(self, association):

	"""Add an association to the list of associations that this topic is an anchor of."""
	
	if not association in self.__associations:
	    self.__associations.append(association)

# =============================================================================

class Name:

    """Name"""
    
    # -----------------------------------------------------------------------------
    # Initialization
    # -----------------------------------------------------------------------------

    def __init__(self, name="", scopes=()):

	self.__name = name
	self.__scopes = list(scopes)

    def __str__(self):
	return self.__name

    # -----------------------------------------------------------------------------
    # Finalizer method
    # -----------------------------------------------------------------------------

    def finalize(self):

	"""Finalize the TopicName object and resolve any missing information."""

	pass
    
    # -----------------------------------------------------------------------------
    # Getters
    # -----------------------------------------------------------------------------

    def get_name(self):

	"""Get name."""
	
	return self.__name

    def get_scopes(self):

	"""Get the scopes that apply to this name."""
	
	return self.__scopes

    # -----------------------------------------------------------------------------
    # Setters
    # -----------------------------------------------------------------------------

    def set_name(self, name):

	"""Set name."""
	
	self.__name = name

    def set_scopes(self, scopes):

	"""Set list of scopes."""

	for scope in scopes:
	    self.add_scope(scope)

    def add_scope(self, scope):

	"""Add scope to the list of scopes."""
	
	self.__scopes.append(scope)

# =============================================================================

class TopicName:

    """5.2.2 Topic Name"""
    
    # -----------------------------------------------------------------------------
    # Initialization
    # -----------------------------------------------------------------------------

    def __init__(self, scopes=(), basenames=(), dispnames=(), sortnames=()):

	self.__scopes = list(scopes)        # Topic+
	self.__basenames = list(basenames)  # Name+
	self.__dispnames = list(dispnames)  # Name+
	self.__sortnames = list(sortnames)  # Name+

    # -----------------------------|------------------------------------------------
    # Finalizer method
    # -----------------------------------------------------------------------------

    def finalize(self):

	"""Finalize the TopicName object and resolve any missing information."""

	pass
    
    # -----------------------------------------------------------------------------
    # Getters
    # -----------------------------------------------------------------------------

    def get_id(self):

	"""Get the internal TopicName id."""
	
	return hash(self)

    def get_scopes(self):

	"""Get the scopes that apply to this topic name."""
	
	return self.__scopes

    def get_basenames(self):

	"""Get the base names of this topic name."""
	
	return self.__basenames

    def get_dispnames(self):

	"""Get the display names of this topic name."""

	return self.__dispnames

    def get_sortnames(self):

	"""Get the sort names/keys of this topic name."""
	
	return self.__sortnames

    # -----------------------------------------------------------------------------
    # Setters
    # -----------------------------------------------------------------------------

    def set_scopes(self, scopes):

	"""Set the scopes that apply to this topic name."""

	for scope in scopes:
	    self.add_scope(scope)

    def add_scope(self, scope):

	"""Add the scoping topic to the list of scoping topics."""
	
	self.__scopes.append(scope)

    def set_basenames(self, basename):

	"""Set the base name of this topic name."""
	
	for basename in basenames:
	    self.add_basename(basename)

    def add_basename(self, basename):

	"""Add the base name to the list of base names."""
	
	self.__basenames.append(basename)

    def set_dispnames(self, dispnames):

	"""Set the display name of this topic name."""

	for dispname in dispnames:
	    self.add_dispname(dispname)

    def add_dispname(self, dispname):

	"""Add the display name to the list of display names."""
	
	self.__dispnames.append(dispname)

    def set_sortnames(self, sortnames):

	"""Set the sort name/key of this topic name."""

	for sortname in sortnames:
	    self.add_sortname(sortname)

    def add_sortname(self, sortname):

	"""Add the sort name to the list of sort names."""
	
	self.__sortnames.append(sortname)

# =============================================================================

class Locator:

    """Location address."""
    
    # -----------------------------------------------------------------------------
    # Initialization
    # -----------------------------------------------------------------------------

    def __init__(self, address="", loctype="sgml"):

	self.__address = address # Locator
	self.__loctype = loctype

    # -----------------------------------------------------------------------------
    # Finalizer method
    # -----------------------------------------------------------------------------

    def finalize(self):

	"""Finalize the Locator object and resolve any missing information."""

	pass
	
    # -----------------------------------------------------------------------------
    # Getters
    # -----------------------------------------------------------------------------

    def get_id(self):

	"""Get the internal Locator id."""
	
	return hash(self)

    def get_address(self):

	"""Get the address of this location."""
	
	return self.__address

    def get_loctype(self):

	"""Get the locator type/notation."""
	
	return self.__loctype
    
    # -----------------------------------------------------------------------------
    # Setters
    # -----------------------------------------------------------------------------

    def set_address(self, address):

	"""Set the address of this location."""
	
	self.__address = address

    def set_loctype(self, loctype):

	"""Set the locator type/notation."""
	
	self.__loctype = loctype

# =============================================================================

class Association:

    """5.3.1 Association link"""
    
    # -----------------------------------------------------------------------------
    # Initialization
    # -----------------------------------------------------------------------------

    def __init__(self, linktype="", type=None, roles=(), scopes=()):

	self.__linktype = linktype    # String
	self.__type = type            # Topic
	self.__roles = list(roles)    # AssociationRole+
	self.__scopes = list(scopes)  # Topic+

    # -----------------------------------------------------------------------------
    # Finalizer method
    # -----------------------------------------------------------------------------

    def finalize(self):

	"""Finalize the Association object and resolve any missing information."""

	pass
	
    # -----------------------------------------------------------------------------
    # Getters
    # -----------------------------------------------------------------------------

    def get_id(self):

	"""Get the internal Association id."""
	
	return hash(self)

    def get_linktype(self):

	"""Get the association link type."""
	
	return self.__linktype

    def get_type(self):

	"""Get the type of this association."""
	
	return self.__type

    def get_roles(self):

	"""Get the association roles of this association."""
	
	return self.__roles

    def get_scopes(self):

	"""Get the scopes that apply to this association."""
	
	return self.__scopes
	
    # -----------------------------------------------------------------------------
    # Setters
    # -----------------------------------------------------------------------------

    def set_linktype(self, linktype):

	"""Set the association link type."""
	
	self.__linktype = linktype

    def set_type(self, type):

	"""Set the association type."""
	
	self.__type = type

    def set_roles(self, roles):

	"""Set the association roles of this association."""

	for role in roles:
	    self.add_role(role)

    def add_role(self, role):

	"""Add the association roles to the list of association roles."""

	self.__roles.append(role)

	# Register association with topics
	for topic in role.get_locators():
	    topic.add_association(self)

    def set_scopes(self, scopes):

	"""Set the scopes that apply to this association."""

	for scope in scopes:
	    self.add_scope(scope)

    def add_scope(self, scope):

	"""Add the scoping topic to the list of scoping topics."""
	
	self.__scopes.append(scope)
	    
# =============================================================================

class AssociationRole:

    """5.3.2 Association role"""
    
    # -----------------------------------------------------------------------------
    # Initialization
    # -----------------------------------------------------------------------------

    def __init__(self, rolename="", type=None, locators=()):

	self.__rolename = rolename        # String
	self.__type = type                # Topic+
	self.__locators = list(locators)  # Topic+/Locator+
	
    # -----------------------------------------------------------------------------
    # Finalizer method
    # -----------------------------------------------------------------------------

    def finalize(self):

	"""Finalize the AssociationRole object and resolve any missing information."""

	pass

    # -----------------------------------------------------------------------------
    # Getters
    # -----------------------------------------------------------------------------

    def get_id(self):

	"""Get the internal AssociationRole id."""
	
	return hash(self)

    def get_rolename(self):

	"""Get the name of this association role."""
	
	return self.__rolename

    def get_type(self):

	"""Get the association type of this association role."""
	
	return self.__type
    
    def get_locators(self):

	"""Get the locators of this association role."""
	
	return self.__locators

    # -----------------------------------------------------------------------------
    # Setters
    # -----------------------------------------------------------------------------

    def set_rolename(self, rolename):

	"""Set the name of this association role."""
	
	self.__rolename = rolename

    def set_type(self, type):

	"""Set the association types of this association role."""

	self.__type = type

    def set_locators(self, locators):

	"""Set the locators of this association role."""

	for locator in locators:
	    self.add_locator(locator)

    def add_locator(self, locator):

	"""Add the locator to the list of locators."""
	
	self.__locators.append(locator)

# =============================================================================

class Facet:

    """5.4.1 Facet link"""

    # -----------------------------------------------------------------------------
    # Initialization
    # -----------------------------------------------------------------------------
    
    def __init__(self, linktype="", type=None, values=()):
	
	self.__linktype = linktype    # String
	self.__type = type            # Topic
	self.__values = list(values)  # FacetValues+
		
    # -----------------------------------------------------------------------------
    # Finalizer method
    # -----------------------------------------------------------------------------

    def finalize(self):

	"""Finalize the Facet object and resolve any missing information."""

	pass

    # -----------------------------------------------------------------------------
    # Getters
    # -----------------------------------------------------------------------------

    def get_id(self):

	"""Get the internal Facet id."""
	
	return hash(self)

    def get_linktype(self):

	"""Get the facet link type."""
	
	return self.__linktype

    def get_type(self):

	"""Get the type of this facet."""
	
	return self.__type

    def get_values(self):

	"""Get the facet values of this facet."""
	
	return self.__values

    # -----------------------------------------------------------------------------
    # Setters
    # -----------------------------------------------------------------------------

    def set_linktype(self, linktype):

	"""Set the facet link type."""
	
	self.__linktype = linktype

    def set_type(self, type):

	"""Set the facet type."""
	
	self.__type = type

    def set_values(self, values):

	"""Set the values of this facet."""

	for value in values:
	    self.add_value(value)

    def add_value(self, value):

	"""Add the value to the list of facet values."""
	
	self.__values.append(value)

# =============================================================================

class FacetValue:

    """5.4.2 Facet value"""
    
    # -----------------------------------------------------------------------------
    # Initialization
    # -----------------------------------------------------------------------------

    def __init__(self, value="", type=None):

	self.__value = value  # String
	self.__type = type    # Topic

	self.__locators = []  # Topic+/Locator+

    # -----------------------------------------------------------------------------
    # Finalizer method
    # -----------------------------------------------------------------------------

    def finalize(self):

	"""Finalize the FacetValue object and resolve any missing information."""

	pass

    # -----------------------------------------------------------------------------
    # Getters
    # -----------------------------------------------------------------------------

    def get_id(self):

	"""Get the internal FacetValue id."""
	
	return hash(self)

    def get_value(self):

	"""Get the value of this facet value."""
	
	return self.__value

    def get_type(self):

	"""Get the type of this facet value."""
	
	return self.__type
    
    def get_locators(self):

	"""Get the locators of this facet value."""
	
	return self.__locators

    # -----------------------------------------------------------------------------
    # Setters
    # -----------------------------------------------------------------------------

    def set_value(self, value):

	"""Set the value of this facet value."""
	
	self.__value = value

    def set_type(self, type):

	"""Set the type of this facet value."""
	
	self.__type = type
    
    def set_locators(self, locators):

	"""Set the locators of this facet value."""
	
	for locator in locators:
	    self.add_locator(locator)

    def add_locator(self, locator):

	"""Add the locator to the list of locators."""
	
	self.__locators.append(locator)

# =============================================================================

class Occurrence:

    """5.2.3 Topic Occurrence"""
    
    # -----------------------------------------------------------------------------
    # Initialization
    # -----------------------------------------------------------------------------

    def __init__(self, rolename="", type=None, scopes=(), locators=()):

	self.__rolename = rolename        # String
	self.__type = type                # Topic
	self.__scopes = list(scopes)      # Topic+
	self.__locators = list(locators)  # Locator+
        self.__value = None               # String

    # -----------------------------------------------------------------------------
    # Finalizer method
    # -----------------------------------------------------------------------------

    def finalize(self):

	"""Finalize the Occurrence object and resolve any missing information."""

	pass

    # -----------------------------------------------------------------------------
    # Getters
    # -----------------------------------------------------------------------------

    def get_id(self):

	"""Get the internal Occurrence id."""
	
	return hash(self)

    def get_rolename(self):

	"""Get the role name of this occurrence."""
	
	return self.__rolename

    def get_type(self):

	"""Get the type of this occurrence."""
	
	return self.__type

    def get_scopes(self):

	"""Get the scopes that apply to this occurrence."""
	
	return self.__scopes
    
    def get_locators(self):

	"""Get the locators of this occurrence."""
	
	return self.__locators

    def get_value(self):

	"""Get the resource that makes up this occurrence."""
	
	return self.__value
    
    # -----------------------------------------------------------------------------
    # Setters
    # -----------------------------------------------------------------------------

    def set_rolename(self, rolename):

	"""Set the role name of this occurrence."""
	
	self.__rolename = rolename

    def set_type(self, type):

	"""Set the type of this occurrence."""
	
	self.__type = type

    def set_scopes(self, scopes):

	"""Set the scoping topics that apply to this occurrence."""

	for scope in scopes:
	    self.add_scope(scope)
    
    def add_scope(self, scope):

	"""Add the scoping topic to the list of scoping topics."""
	
	self.__scopes.append(scope)

    def set_locators(self, locators):

	"""Set the locators of this occurrence."""

	for locator in locators:
	    self.add_locator(locator)

    def add_locator(self, locator):

	"""Add the locator to the list of locators."""
	
	self.__locators.append(locator)

    def set_value(self, value):

	"""Set the resource that makes up this occurrence."""
	
	self.__value = value
        
# ============================================================================
# Functions
# ============================================================================

def dict_has(dict, method, args, cmp_item):
    
    """Checks whether cmp_item is the result of applying method to keys of the dict."""

    for key in dict.keys():
	i = apply(getattr(key, method), args)
	if i == cmp_item:
	    return key

def remove_from_seq(seq, item, msg=None):

    """Removes all occurrences of item in the sequence. Throws an exception if the key doesn't exist."""

    # Remove first
    try:
	while 1:
	    del seq[seq.index(item)]
    except ValueError,e:
	if msg:
	    raise msg
	return

    # Remove rest
    try:
	while 1:
	    del seq[seq.index(item)]
    except ValueError,e:
	pass

def remove_from_dict(dict, key, msg=None):

    """Removes an item from the dictionary. Throws an exception if the key doesn't exist."""

    try:
	del dict[key]
    except:
	if msg:
	    raise msg

# ============================================================================
# Debug level
# ============================================================================

debug = 0
						       
def set_debug(level):

    """Set the debug level."""
    
    debug = level
